"""helper pyteal functions"""
from pyteal import (App, Global, TxnType, Seq, And, TealType, Txn, Int, Expr,
                    Gtxn, If, Bytes, InnerTxnBuilder, TxnField, InnerTxn, ScratchVar, AssetHolding)

from amm.contracts.config import (
    POOL_TOKENS_OUTSTANDING_KEY, POOL_TOKEN_KEY, POOL_FUNDING_RESERVES, RESULT,
    YES_TOKEN_KEY, YES_TOKENS_OUTSTANDING_KEY, YES_TOKENS_RESERVES,
    NO_TOKEN_KEY, NO_TOKENS_OUTSTANDING_KEY, NO_TOKENS_RESERVES,
    TOKEN_FUNDING_KEY, TOKEN_FUNDING_RESERVES
)


def validate_token_received(
    transaction_index: TealType.uint64, token_key: TealType.bytes
) -> Expr:
    """validate token type, sender, asset """
    return And(
        Gtxn[transaction_index].type_enum() == TxnType.AssetTransfer,
        Gtxn[transaction_index].sender() == Txn.sender(),
        Gtxn[transaction_index].asset_receiver()
        == Global.current_application_address(),
        Gtxn[transaction_index].xfer_asset() == App.globalGet(token_key),
        Gtxn[transaction_index].asset_amount() > Int(0),
    )


def send_token(
    token_key: TealType.bytes, receiver: TealType.bytes, amount: TealType.uint64
) -> Expr:
    """asa token transfer"""
    return Seq(
        InnerTxnBuilder.Begin(),
        InnerTxnBuilder.SetFields(
            {
                TxnField.type_enum: TxnType.AssetTransfer,
                TxnField.xfer_asset: App.globalGet(token_key),
                TxnField.asset_receiver: receiver,
                TxnField.asset_amount: amount,
            }
        ),
        InnerTxnBuilder.Submit(),
    )


def opt_in(token_key: TealType.bytes) -> Expr:
    """asa opt in"""
    return send_token(token_key, Global.current_application_address(), Int(0))


def create_pool_token(pool_token_amount: TealType.uint64) -> Expr:
    """creates asa token"""
    return Seq(
        InnerTxnBuilder.Begin(),
        InnerTxnBuilder.SetFields(
            {
                TxnField.type_enum: TxnType.AssetConfig,
                TxnField.config_asset_total: pool_token_amount,
                TxnField.config_asset_name: Bytes("PoolToken"),
                TxnField.config_asset_default_frozen: Int(0),
                TxnField.config_asset_decimals: Int(0),
                TxnField.config_asset_reserve: Global.current_application_address(),
            }
        ),
        InnerTxnBuilder.Submit(),
        App.globalPut(POOL_TOKEN_KEY, InnerTxn.created_asset_id()),
        App.globalPut(POOL_TOKENS_OUTSTANDING_KEY, Int(0)),
    )


def create_no_token(token_amount: TealType.uint64) -> Expr:
    """creates asa token"""
    return Seq(
        InnerTxnBuilder.Begin(),
        InnerTxnBuilder.SetFields(
            {
                TxnField.type_enum: TxnType.AssetConfig,
                TxnField.config_asset_total: token_amount,
                TxnField.config_asset_name: Bytes("NoToken"),
                TxnField.config_asset_unit_name: Bytes("No"),
                TxnField.config_asset_default_frozen: Int(0),
                TxnField.config_asset_decimals: Int(0),
                TxnField.config_asset_reserve: Global.current_application_address(),
            }
        ),
        InnerTxnBuilder.Submit(),
        App.globalPut(NO_TOKEN_KEY, InnerTxn.created_asset_id()),
        App.globalPut(NO_TOKENS_OUTSTANDING_KEY, Int(0)),
        App.globalPut(NO_TOKENS_RESERVES, Int(0)),
    )


def create_yes_token(token_amount: TealType.uint64) -> Expr:
    """creates asa token"""
    return Seq(
        InnerTxnBuilder.Begin(),
        InnerTxnBuilder.SetFields(
            {
                TxnField.type_enum: TxnType.AssetConfig,
                TxnField.config_asset_total: token_amount,
                TxnField.config_asset_name: Bytes("YesToken"),
                TxnField.config_asset_unit_name: Bytes("Yes"),
                TxnField.config_asset_default_frozen: Int(0),
                TxnField.config_asset_decimals: Int(0),
                TxnField.config_asset_reserve: Global.current_application_address(),
            }
        ),
        InnerTxnBuilder.Submit(),
        App.globalPut(YES_TOKEN_KEY, InnerTxn.created_asset_id()),
        App.globalPut(YES_TOKENS_OUTSTANDING_KEY, Int(0)),
        App.globalPut(YES_TOKENS_RESERVES, Int(0)),
    )


def mint_and_send_pool_token(receiver: TealType.bytes, amount: TealType.uint64) -> Expr:
    """mintAndSendPoolToken"""
    ratio: ScratchVar = ScratchVar(TealType.uint64)
    no_reserves = App.globalGet(NO_TOKENS_RESERVES)
    return Seq(
        ratio.store(
            (Int(1) + App.globalGet(NO_TOKENS_RESERVES)) /
            (Int(1) + App.globalGet(YES_TOKENS_RESERVES))),
        If(App.globalGet(POOL_FUNDING_RESERVES) > Int(0)).Then(
            Seq(
                send_token(
                    POOL_TOKEN_KEY, receiver, amount * App.globalGet(POOL_TOKENS_OUTSTANDING_KEY) /
                    App.globalGet(POOL_FUNDING_RESERVES)),
                App.globalPut(
                    POOL_TOKENS_OUTSTANDING_KEY, App.globalGet(
                        POOL_TOKENS_OUTSTANDING_KEY) + amount
                    * App.globalGet(POOL_TOKENS_OUTSTANDING_KEY) / App.globalGet(
                        POOL_FUNDING_RESERVES)))).ElseIf(
            App.globalGet(POOL_FUNDING_RESERVES) == Int(0)).Then(
            Seq(
                send_token(POOL_TOKEN_KEY, receiver, amount),
                App.globalPut(
                    POOL_TOKENS_OUTSTANDING_KEY, App.globalGet(
                        POOL_TOKENS_OUTSTANDING_KEY)
                    + amount,),)),
        App.globalPut(
            NO_TOKENS_RESERVES, ratio.load() * (amount / Int(4)) + no_reserves),
        App.globalPut(
            YES_TOKENS_RESERVES, (Int(1) / ratio.load()) * (amount / Int(4)) + App.globalGet(
                YES_TOKENS_RESERVES)),
        App.globalPut(POOL_FUNDING_RESERVES, App.globalGet(POOL_FUNDING_RESERVES) + amount),)


def mint_and_send_no_token(
    receiver: TealType.bytes, amount: TealType.uint64
) -> Expr:
    """mints no token"""
    funding = AssetHolding.balance(
        Global.current_application_address(), App.globalGet(TOKEN_FUNDING_KEY)
    )
    tokens_out: ScratchVar = ScratchVar(TealType.uint64)
    return Seq(
        tokens_out.store(
            App.globalGet(NO_TOKENS_RESERVES) * amount /
            (App.globalGet(YES_TOKENS_RESERVES) + amount)
        ),
        App.globalPut(NO_TOKENS_OUTSTANDING_KEY,
                      App.globalGet(NO_TOKENS_OUTSTANDING_KEY) + tokens_out.load()),
        App.globalPut(NO_TOKENS_RESERVES, App.globalGet(
            NO_TOKENS_RESERVES) - tokens_out.load()),
        send_token(NO_TOKEN_KEY, receiver, tokens_out.load()),
        If(App.globalGet(NO_TOKENS_OUTSTANDING_KEY) >
           App.globalGet(YES_TOKENS_OUTSTANDING_KEY))
        .Then(
            App.globalPut(TOKEN_FUNDING_RESERVES, App.globalGet(
                NO_TOKENS_OUTSTANDING_KEY) * Int(2))
        ),
        funding,
        App.globalPut(
            POOL_FUNDING_RESERVES,
            funding.value() - App.globalGet(TOKEN_FUNDING_RESERVES)
        ),
    )


def mint_and_send_yes_token(
    receiver: TealType.bytes, amount: TealType.uint64
) -> Expr:
    """mints yes token"""
    funding = AssetHolding.balance(
        Global.current_application_address(), App.globalGet(TOKEN_FUNDING_KEY)
    )
    tokens_out: ScratchVar = ScratchVar(TealType.uint64)
    return Seq(
        tokens_out.store(
            (App.globalGet(YES_TOKENS_RESERVES) * amount /
             (App.globalGet(NO_TOKENS_RESERVES) + amount))
        ),
        App.globalPut(YES_TOKENS_OUTSTANDING_KEY,
                      App.globalGet(YES_TOKENS_OUTSTANDING_KEY) + tokens_out.load()),
        App.globalPut(YES_TOKENS_RESERVES, App.globalGet(
            YES_TOKENS_RESERVES) - tokens_out.load()),
        send_token(YES_TOKEN_KEY, receiver, tokens_out.load()),
        If(App.globalGet(YES_TOKENS_OUTSTANDING_KEY)
           > App.globalGet(NO_TOKENS_OUTSTANDING_KEY))
        .Then(
            App.globalPut(TOKEN_FUNDING_RESERVES,
                          App.globalGet(YES_TOKENS_OUTSTANDING_KEY) * Int(2)
                          )
        ),
        funding,
        App.globalPut(
            POOL_FUNDING_RESERVES,
            funding.value() - App.globalGet(TOKEN_FUNDING_RESERVES)
        ),
    )


def withdraw_lp_token(
    receiver: TealType.bytes,
    pool_token_amount: TealType.uint64,
) -> Expr:
    """withdraws lp token"""
    ratio: ScratchVar = ScratchVar(TealType.uint64)
    total = App.globalGet(POOL_TOKENS_OUTSTANDING_KEY)
    reserves = App.globalGet(POOL_FUNDING_RESERVES)

    return Seq(
        send_token(
            TOKEN_FUNDING_KEY,
            receiver,
            App.globalGet(POOL_FUNDING_RESERVES) *
            pool_token_amount / App.globalGet(POOL_TOKENS_OUTSTANDING_KEY),
        ),
        App.globalPut(
            POOL_FUNDING_RESERVES,
            App.globalGet(POOL_FUNDING_RESERVES) - (App.globalGet(POOL_FUNDING_RESERVES) *
                                                    pool_token_amount / total),
        ),
        App.globalPut(
            POOL_TOKENS_OUTSTANDING_KEY,
            App.globalGet(POOL_TOKENS_OUTSTANDING_KEY) - pool_token_amount,
        ),
        If(App.globalGet(RESULT) == Int(0)).Then(
            Seq(
                ratio.store(
                    (Int(1) + App.globalGet(NO_TOKENS_RESERVES))
                    / (Int(1) + App.globalGet(YES_TOKENS_RESERVES))
                ),
                App.globalPut(NO_TOKENS_RESERVES,
                              App.globalGet(NO_TOKENS_RESERVES) - (reserves *
                                                                   pool_token_amount / total) /
                              Int(4) * ratio.load()),
                App.globalPut(YES_TOKENS_RESERVES,
                              App.globalGet(YES_TOKENS_RESERVES) - (reserves *
                                                                    pool_token_amount / total) /
                              Int(4) * (Int(1)/ratio.load())),
            )),
    )


def redeem_token(
    receiver: TealType.bytes,
    result_token_amount: TealType.uint64,
) -> Expr:
    """redeems token"""
    return Seq(
        send_token(
            TOKEN_FUNDING_KEY,
            receiver,
            result_token_amount * Int(2)
        ),
        App.globalPut(
            TOKEN_FUNDING_RESERVES,
            App.globalGet(TOKEN_FUNDING_RESERVES) -
            result_token_amount * Int(2)
        ),
        If(App.globalGet(RESULT) == App.globalGet(YES_TOKEN_KEY))
        .Then(
            App.globalPut(
                YES_TOKENS_OUTSTANDING_KEY,
                App.globalGet(YES_TOKENS_OUTSTANDING_KEY) - result_token_amount
            ),
        )
        .ElseIf(App.globalGet(RESULT) == App.globalGet(NO_TOKEN_KEY))
        .Then(
            App.globalPut(
                NO_TOKENS_OUTSTANDING_KEY,
                App.globalGet(NO_TOKENS_OUTSTANDING_KEY) - result_token_amount
            ),
        ),
    )



"""tests"""
from unittest import TestCase
import os
from base64 import b64decode
from dotenv import load_dotenv

from amm.amm_app import App
from amm.utils.account import Account
from amm.utils.purestake_client import AlgoClient

load_dotenv()


class TestAmm(TestCase):
    """Class for testing lifecycle of the amm"""

    def setUp(self):
        self.algod_token = os.getenv('algod_token')
        self.deployer = Account(os.getenv('key'))
        self.algo_client = AlgoClient(self.algod_token)
        self.app = App(self.algo_client.client)

    def test_is_client(self):
        """tests environment variables and client"""
        assert isinstance(self.algod_token, str),  "Provide Algorand node key"
        assert isinstance(self.deployer.private_key,
                          str),  "Provide deployer private key"

    def test_create_amm(self):
        """tests amm"""
        stable_token = self.algo_client.create_asset(self.deployer)
        assert isinstance(
            stable_token, int), "Provide sufficient algo to deployer"

        app_id = self.app.create_amm_app(
            token=stable_token,
            min_increment=1000,
            deployer=self.deployer
        )

        assert isinstance(app_id, int), "Provide sufficient algo to deployer"

        pool_tokens = self.app.setup_amm_app(
            funder=self.deployer
        )
        pool_token = pool_tokens['pool_token_key']
        yes_token = pool_tokens['yes_token_key']
        no_token = pool_tokens['no_token_key']

        assert isinstance(pool_token, int), "failed to create pool token"
        assert isinstance(yes_token, int), "failed to create yes token"
        assert isinstance(no_token, int), "failed to create no token"

        self.app.opt_in_to_pool_token(self.deployer)
        self.app.opt_in_to_yes_token(self.deployer)
        self.app.opt_in_to_no_token(self.deployer)

        pool_token_amount = 2_000_000

        self.app.supply(
            quantity=pool_token_amount,
            supplier=self.deployer,
        )

        swap_token_amount = 100_000
        self.app.swap(
            option="yes",
            quantity=swap_token_amount,
            supplier=self.deployer
        )

        self.app.swap(
            option="no",
            quantity=swap_token_amount,
            supplier=self.deployer
        )

        self.app.set_result(
            second_argument=b"yes",
            funder=self.deployer
        )

        app = self.algo_client.client.application_info(app_id)

        glob_state = app['params']['global-state']

        ids = {}
        key = ""
        for i, _ in enumerate(glob_state):
            key = b64decode(glob_state[i]['key']).decode("utf-8")
            if glob_state[i]['value']['uint'] != 0:
                ids[key] = glob_state[i]['value']['uint']
            else:
                ids[key] = glob_state[i]['value']['bytes']

        print(ids)

        yes_token_amount = 83_333

        self.app.redeem(
            token_in=yes_token,
            token_amount=yes_token_amount,
            withdrawal_account=self.deployer,
            token_out=stable_token
        )

        self.app.withdraw(
            pool_token_amount=pool_token_amount, withdrawal_account=self.deployer
        )

        self.app.close_amm(
            closing_account=self.deployer
        )



"""api to the contract"""
from typing import Tuple
from base64 import b64decode

from pyteal import compileTeal, Mode, Expr
from algosdk import encoding
from algosdk.transaction import (StateSchema, ApplicationCreateTxn,
                                 OnComplete, PaymentTxn, ApplicationCallTxn, assign_group_id,
                                 AssetOptInTxn, AssetTransferTxn, ApplicationDeleteTxn)
from algosdk.v2client.algod import AlgodClient
from algosdk.logic import get_application_address

from amm.utils.account import Account
from amm.contracts.amm import approval_program, clear_program

MIN_BALANCE_REQUIREMENT = (
    # min account balance
    110_000
    # additional min balance for 4 assets
    + 100_000 * 4
)


def fully_compile_contract(
    client: AlgodClient, contract: Expr
) -> bytes:
    """
    Compiles teal.
    Args:
        client: algorand client
        contract: teal contract
    Returns: bytecode
    """
    teal = compileTeal(contract, mode=Mode.Application, version=6)
    response = client.compile(teal)

    return b64decode(response["result"])


def get_contracts(client: AlgodClient) -> Tuple[bytes, bytes]:
    """
    Get the compiled TEAL contracts for the AMM.
    Args:
        client: An algod client that has the ability to compile TEAL programs.
    Returns:
        The approval program and the clear state program.
    """

    approval_program_compiled = fully_compile_contract(
        client, approval_program())
    clear_state_program_compiled = fully_compile_contract(
        client, clear_program())

    return approval_program_compiled, clear_state_program_compiled


class App:
    """ Algorand App """

    def __init__(self, client: AlgodClient, app_id=0):
        self.client = client
        self.app_id = app_id
        self.app_addr = get_application_address(app_id)
        self.stable_token: int
        self.pool_token: int
        self.yes_token: int
        self.no_token: int

    def update_app_address(self):
        """
        Updates app address.
        """
        self.app_addr = get_application_address(self.app_id)

    def wait_for_transaction(
        self, tx_id: str, timeout: int = 10
    ):
        """
        Monitors transaction completion.
        Args:
            tx_id: transaction id
            timeout: timeout in rounds
        """

        last_status = self.client.status()
        last_round = last_status["last-round"]
        start_round = last_round

        while last_round < start_round + timeout:
            pending_txn = self.client.pending_transaction_info(tx_id)

            if pending_txn.get("confirmed-round", 0) > 0:
                return pending_txn

            if pending_txn["pool-error"]:
                raise RuntimeError("Pool error:")

            last_status = self.client.status_after_block(last_round + 1)

            last_round += 1

        raise RuntimeError(
            f"Transaction {tx_id} not confirmed after {timeout} rounds"
        )

    def create_amm_app(
        self,
        token: int,
        min_increment: int,
        deployer: Account
    ) -> int:
        """
        Creates a new amm.
        Args:
            deployer: The account that will create the amm application.
            token: The id of liquidity token in the liquidity pool,
            min_increment: min int to fund the pool
        Return:
            The ID of the newly created amm app.
        """
        self.stable_token = token
        approval, clear = get_contracts(self.client)

        global_schema = StateSchema(
            num_uints=13, num_byte_slices=1)
        local_schema = StateSchema(num_uints=0, num_byte_slices=0)

        app_args = [
            encoding.decode_address(deployer.public_key),
            token.to_bytes(8, "big"),
            min_increment.to_bytes(8, "big"),
        ]

        txn = ApplicationCreateTxn(
            sender=deployer.public_key,
            on_complete=OnComplete.NoOpOC,
            approval_program=approval,
            clear_program=clear,
            global_schema=global_schema,
            local_schema=local_schema,
            app_args=app_args,
            sp=self.client.suggested_params(),
        )

        signed_tx = txn.sign(deployer.private_key)

        self.client.send_transaction(signed_tx)

        response = self.wait_for_transaction(signed_tx.get_txid())
        assert response["application-index"] is not None and response["application-index"] > 0
        self.app_id = response["application-index"]
        self.update_app_address()
        return response["application-index"]

    def setup_amm_app(
        self,
        funder: Account
    ) -> int:
        """
        Finish setting up an amm.
        This operation funds the pool account, creates pool token,
        and opts app into tokens A and B, all in one atomic transaction group.
        Args:
            funder: The account providing the funding for the escrow account.
        Return: app asset ids.
        """

        fund_app_tx = PaymentTxn(
            sender=funder.public_key,
            receiver=self.app_addr,
            amt=MIN_BALANCE_REQUIREMENT,
            sp=self.client.suggested_params(),
        )

        setup_tx = ApplicationCallTxn(
            sender=funder.public_key,
            index=self.app_id,
            on_complete=OnComplete.NoOpOC,
            app_args=[b"setup"],
            foreign_assets=[self.stable_token],
            sp=self.client.suggested_params(),
        )

        assign_group_id([fund_app_tx, setup_tx])

        signed_fund_spp_txn = fund_app_tx.sign(funder.private_key)
        signed_setup_tx = setup_tx.sign(funder.private_key)

        self.client.send_transactions([signed_fund_spp_txn, signed_setup_tx])

        self.wait_for_transaction(signed_fund_spp_txn.get_txid())
        glob_state = self.client.application_info(
            self.app_id)['params']['global-state']

        ids = {}

        for i, _ in enumerate(glob_state):
            if b64decode(glob_state[i]['key']) == b"pool_token_key":
                ids['pool_token_key'] = glob_state[i]['value']['uint']
                self.pool_token = glob_state[i]['value']['uint']
            elif b64decode(glob_state[i]['key']) == b"yes_token_key":
                ids['yes_token_key'] = glob_state[i]['value']['uint']
                self.yes_token = glob_state[i]['value']['uint']
            elif b64decode(glob_state[i]['key']) == b"no_token_key":
                ids['no_token_key'] = glob_state[i]['value']['uint']
                self.no_token = glob_state[i]['value']['uint']

        return ids

    def opt_in_to_pool_token(
        self,
        account: Account
    ) -> None:
        """
        Opts into Pool Token.
        Args:
            account: The account opting into the token.
        """

        opt_in_tx = AssetOptInTxn(
            sender=account.public_key, index=self.pool_token, sp=self.client.suggested_params()
        )

        signed_opt_in_tx = opt_in_tx.sign(account.private_key)

        self.client.send_transaction(signed_opt_in_tx)
        self.wait_for_transaction(signed_opt_in_tx.get_txid())

    def opt_in_to_no_token(
        self,
        account: Account
    ) -> None:
        """
        Opts into Pool Token.
        Args:
            account: The account opting into the token.
        """

        opt_in_tx = AssetOptInTxn(
            sender=account.public_key, index=self.no_token, sp=self.client.suggested_params()
        )

        signed_opt_in_tx = opt_in_tx.sign(account.private_key)

        self.client.send_transaction(signed_opt_in_tx)
        self.wait_for_transaction(signed_opt_in_tx.get_txid())

    def opt_in_to_yes_token(
        self,
        account: Account
    ) -> None:
        """
        Opts into Pool Token.
        Args:
            account: The account opting into the token.
        """

        opt_in_tx = AssetOptInTxn(
            sender=account.public_key, index=self.yes_token, sp=self.client.suggested_params()
        )

        signed_opt_in_tx = opt_in_tx.sign(account.private_key)

        self.client.send_transaction(signed_opt_in_tx)
        self.wait_for_transaction(signed_opt_in_tx.get_txid())

    def supply(
        self, quantity: int, supplier: Account
    ) -> None:
        """
        Supply liquidity to the pool.
        Args:
            quantity: quantity
            supplier: supplier
        """

        fee_tx = PaymentTxn(
            sender=supplier.public_key,
            receiver=self.app_addr,
            amt=MIN_BALANCE_REQUIREMENT,
            sp=self.client.suggested_params(),
        )

        token_tx = AssetTransferTxn(
            sender=supplier.public_key,
            receiver=self.app_addr,
            index=self.stable_token,
            amt=quantity,
            sp=self.client.suggested_params(),
        )

        app_call_tx = ApplicationCallTxn(
            sender=supplier.public_key,
            index=self.app_id,
            on_complete=OnComplete.NoOpOC,
            app_args=[b"supply"],
            foreign_assets=[self.stable_token, self.pool_token,
                            self.yes_token, self.no_token],
            sp=self.client.suggested_params(),
        )

        assign_group_id([fee_tx, token_tx, app_call_tx])
        signed_fee_tx = fee_tx.sign(supplier.private_key)
        signed_token_tx = token_tx.sign(supplier.private_key)
        signed_app_call_tx = app_call_tx.sign(supplier.private_key)

        self.client.send_transactions(
            [signed_fee_tx, signed_token_tx, signed_app_call_tx]
        )
        self.wait_for_transaction(signed_app_call_tx.get_txid())

    def swap(
        self, option: str, quantity: int, supplier: Account
    ) -> None:
        """
        Swap stable for an AMM option.
        Args:
            option: string either yes or no option
            quantity: token amount
            supplier: account of the sender
        """
        if option == 'yes':
            second_argument = b"buy_yes"
        elif option == 'no':
            second_argument = b"buy_no"
        else:
            return

        fee_tx = PaymentTxn(
            sender=supplier.public_key,
            receiver=self.app_addr,
            amt=2_000,
            sp=self.client.suggested_params(),
        )

        token_tx = AssetTransferTxn(
            sender=supplier.public_key,
            receiver=self.app_addr,
            index=self.stable_token,
            amt=quantity,
            sp=self.client.suggested_params(),
        )

        app_call_tx = ApplicationCallTxn(
            sender=supplier.public_key,
            index=self.app_id,
            on_complete=OnComplete.NoOpOC,
            app_args=[b"swap", second_argument],
            foreign_assets=[self.stable_token, self.pool_token,
                            self.yes_token, self.no_token],
            sp=self.client.suggested_params(),
        )

        assign_group_id([fee_tx, token_tx, app_call_tx])
        signed_fee_tx = fee_tx.sign(supplier.private_key)
        signed_token_tx = token_tx.sign(supplier.private_key)
        signed_app_call_tx = app_call_tx.sign(supplier.private_key)

        self.client.send_transactions(
            [signed_fee_tx, signed_token_tx, signed_app_call_tx]
        )
        self.wait_for_transaction(signed_app_call_tx.get_txid())

    def withdraw(
        self, pool_token_amount: int,
        withdrawal_account: Account
    ) -> None:
        """
        Withdraw liquidity  + rewards from the pool back to supplier.
        Supplier should receive stablecoin + fees proportional
        to the liquidity share in the pool they choose to withdraw.
        Args:
            pool_token_amount: token amount
            withdrawal_account: account of the sender to withdraw stablecoins
        """

        fee_tx = PaymentTxn(
            sender=withdrawal_account.public_key,
            receiver=self.app_addr,
            amt=2_000,
            sp=self.client.suggested_params(),
        )

        pool_token_tx = AssetTransferTxn(
            sender=withdrawal_account.public_key,
            receiver=self.app_addr,
            index=self.pool_token,
            amt=pool_token_amount,
            sp=self.client.suggested_params(),
        )

        app_call_tx = ApplicationCallTxn(
            sender=withdrawal_account.public_key,
            index=self.app_id,
            on_complete=OnComplete.NoOpOC,
            app_args=[b"withdraw"],
            foreign_assets=[self.stable_token, self.pool_token],
            sp=self.client.suggested_params(),
        )

        assign_group_id([fee_tx, pool_token_tx, app_call_tx])
        signed_fee_tx = fee_tx.sign(withdrawal_account.private_key)
        signed_token_tx = pool_token_tx.sign(withdrawal_account.private_key)
        signed_app_call_tx = app_call_tx.sign(withdrawal_account.private_key)

        self.client.send_transactions(
            [signed_fee_tx, signed_token_tx, signed_app_call_tx])
        self.wait_for_transaction(signed_app_call_tx.get_txid())

    def redeem(
        self, token_in: int, token_amount: int,
        withdrawal_account: Account, token_out: int
    ) -> None:
        """
        Redeems winning token for stablecoin.
        First, pay for the fee incurred by AMM for sending back tokens A and B.
        Then conduct asset transfer and application call.
        Args:
            token_in: should be a winning token
            token_amount: token amount
            withdrawal_account: withdrawal account
            token_out: stablecoin
        """

        fee_tx = PaymentTxn(
            sender=withdrawal_account.public_key,
            receiver=self.app_addr,
            amt=2_000,
            sp=self.client.suggested_params(),
        )

        token_tx = AssetTransferTxn(
            sender=withdrawal_account.public_key,
            receiver=self.app_addr,
            index=token_in,
            amt=token_amount,
            sp=self.client.suggested_params(),
        )

        app_call_tx = ApplicationCallTxn(
            sender=withdrawal_account.public_key,
            index=self.app_id,
            on_complete=OnComplete.NoOpOC,
            app_args=[b"redeem"],
            foreign_assets=[token_out, token_in],
            sp=self.client.suggested_params(),
        )

        assign_group_id([fee_tx, token_tx, app_call_tx])
        signed_fee_tx = fee_tx.sign(withdrawal_account.private_key)
        signed_token_tx = token_tx.sign(withdrawal_account.private_key)
        signed_app_call_tx = app_call_tx.sign(withdrawal_account.private_key)

        self.client.send_transactions(
            [signed_fee_tx, signed_token_tx, signed_app_call_tx])
        self.wait_for_transaction(signed_app_call_tx.get_txid())

    def set_result(
        self,
        funder: Account,
        second_argument
    ) -> None:
        """
        Sets result of the event.
        Args:
            funder: creator of the amm
            second_argument: result
        """

        fee_tx = PaymentTxn(
            sender=funder.public_key,
            receiver=self.app_addr,
            amt=2_000,
            sp=self.client.suggested_params(),
        )

        call_tx = ApplicationCallTxn(
            sender=funder.public_key,
            index=self.app_id,
            on_complete=OnComplete.NoOpOC,
            app_args=[b"result", second_argument],
            sp=self.client.suggested_params(),
        )

        assign_group_id([fee_tx, call_tx])
        signed_fee_tx = fee_tx.sign(funder.private_key)
        signed_app_call_tx = call_tx.sign(funder.private_key)

        self.client.send_transactions([signed_fee_tx, signed_app_call_tx])
        self.wait_for_transaction(signed_app_call_tx.get_txid())

    def close_amm(
        self,
        closing_account: Account
    ) -> None:
        """
        Closes an AMM.
        Args:
            closing_account: closer account public address.
            Must be the original creator of the pool.
        """

        delete_tx = ApplicationDeleteTxn(
            sender=closing_account.public_key,
            index=self.app_id,
            sp=self.client.suggested_params(),
        )
        signed_tx = delete_tx.sign(closing_account.private_key)

        self.client.send_transaction(signed_tx)

        self.wait_for_transaction(signed_tx.get_txid())



"""main"""
from pyteal import (App, Global, Assert, Seq, And, Not, Txn, Int,
                    Approve, Gtxn, If, Bytes, Reject, Btoi, Cond, Or, OnComplete, compileTeal, Mode)

from amm.contracts.helpers import (
    validate_token_received, mint_and_send_pool_token,
    mint_and_send_no_token, mint_and_send_yes_token,
    opt_in, create_pool_token, withdraw_lp_token,
    create_no_token, create_yes_token, redeem_token
)

from amm.contracts.config import (
    CREATOR_KEY, TOKEN_FUNDING_KEY,
    POOL_TOKEN_KEY, MIN_INCREMENT_KEY,
    POOL_TOKENS_OUTSTANDING_KEY, TOKEN_DEFAULT_AMOUNT,
    YES_TOKEN_KEY, NO_TOKEN_KEY,
    TOKEN_FUNDING_RESERVES, POOL_FUNDING_RESERVES,
    RESULT
)


def get_setup():
    """sets up contract"""
    pool_token_id = App.globalGetEx(
        Global.current_application_id(), POOL_TOKEN_KEY)
    pool_tokens_outstanding = App.globalGetEx(
        Global.current_application_id(), POOL_TOKENS_OUTSTANDING_KEY
    )
    on_setup = Seq(
        pool_token_id,
        pool_tokens_outstanding,
        Assert(Not(pool_token_id.hasValue())),
        Assert(Not(pool_tokens_outstanding.hasValue())),
        create_pool_token(TOKEN_DEFAULT_AMOUNT),
        opt_in(TOKEN_FUNDING_KEY),
        create_no_token(TOKEN_DEFAULT_AMOUNT),
        opt_in(NO_TOKEN_KEY),
        create_yes_token(TOKEN_DEFAULT_AMOUNT),
        opt_in(YES_TOKEN_KEY),
        Approve(),
    )
    return on_setup


def get_supply():
    """liquidity supply"""
    token_txn_index = Txn.group_index() - Int(1)

    on_supply = Seq(
        Assert(
            And(
                validate_token_received(token_txn_index, TOKEN_FUNDING_KEY),
                Gtxn[token_txn_index].asset_amount()
                >= App.globalGet(MIN_INCREMENT_KEY),
            )
        ),
        mint_and_send_pool_token(
            Txn.sender(),
            Gtxn[token_txn_index].asset_amount(),
        ),
        Approve(),
    )
    return on_supply


def get_swap():
    """option swap"""
    token_txn_index = Txn.group_index() - Int(1)
    option = Txn.application_args[1]
    on_swap = Seq(
        Assert(
            validate_token_received(token_txn_index, TOKEN_FUNDING_KEY),
        ),
        If(option == Bytes("buy_yes"))
        .Then(
            Seq(
                mint_and_send_yes_token(
                    Txn.sender(),
                    Gtxn[token_txn_index].asset_amount(),
                ),
                Approve()
            ),
        )
        .ElseIf(option == Bytes("buy_no"))
        .Then(
            Seq(
                mint_and_send_no_token(
                    Txn.sender(),
                    Gtxn[token_txn_index].asset_amount(),
                ),
                Approve()
            ),
        ),
        Reject()
    )

    return on_swap


def get_withdraw():
    """liquidity withdrawal"""
    pool_token_txn_index = Txn.group_index() - Int(1)

    on_withdraw = Seq(
        Assert(
            validate_token_received(pool_token_txn_index, POOL_TOKEN_KEY),
        ),
        withdraw_lp_token(
            Txn.sender(),
            Gtxn[pool_token_txn_index].asset_amount(),
        ),
        Approve(),
    )

    return on_withdraw


def get_result():
    """sets result"""
    result = Txn.application_args[1]

    on_result = Seq(
        Assert(
            Txn.sender() == App.globalGet(CREATOR_KEY)
        ),

        If(result == Bytes("yes"))
        .Then(
            Seq(
                App.globalPut(RESULT, App.globalGet(YES_TOKEN_KEY)),
                Approve()
            )
        )
        .ElseIf(
            result == Bytes("no"),
        )
        .Then(
            Seq(
                App.globalPut(RESULT, App.globalGet(NO_TOKEN_KEY)),
                Approve()
            )
        ),
        Reject(),
    )

    return on_result


def get_redemption():
    """redeems winning tokens"""
    token_txn_index = Txn.group_index() - Int(1)

    on_redemption = Seq(
        Assert(
            And(
                validate_token_received(token_txn_index, RESULT),
            )
        ),
        redeem_token(
            Txn.sender(),
            Gtxn[token_txn_index].asset_amount(),
        ),
        Approve(),
    )

    return on_redemption


def approval_program():
    """main"""
    on_create = Seq(
        App.globalPut(CREATOR_KEY, Txn.application_args[0]),
        App.globalPut(TOKEN_FUNDING_KEY, Btoi(Txn.application_args[1])),
        App.globalPut(MIN_INCREMENT_KEY, Btoi(Txn.application_args[2])),
        App.globalPut(TOKEN_FUNDING_RESERVES, Int(0)),
        App.globalPut(POOL_FUNDING_RESERVES, Int(0)),
        App.globalPut(RESULT, Int(0)),
        Approve(),
    )

    on_setup = get_setup()
    on_supply = get_supply()
    on_swap = get_swap()
    on_withdraw = get_withdraw()
    on_redemption = get_redemption()
    on_result = get_result()

    on_call_method = Txn.application_args[0]
    on_call = Cond(
        [on_call_method == Bytes("setup"), on_setup],
        [on_call_method == Bytes("supply"), on_supply],
        [on_call_method == Bytes("withdraw"), on_withdraw],
        [on_call_method == Bytes("swap"), on_swap],
        [on_call_method == Bytes("redeem"), on_redemption],
        [on_call_method == Bytes("result"), on_result],
    )

    on_delete = Seq(
        Assert(
            And(
                Txn.sender() == App.globalGet(CREATOR_KEY),
                App.globalGet(POOL_TOKENS_OUTSTANDING_KEY) == Int(0),
            ),
        ),
        Approve(),
    )

    program = Cond(
        [Txn.application_id() == Int(0), on_create],
        [Txn.on_completion() == OnComplete.NoOp, on_call],
        [Txn.on_completion() == OnComplete.DeleteApplication, on_delete],
        [
            Or(
                Txn.on_completion() == OnComplete.OptIn,
                Txn.on_completion() == OnComplete.CloseOut,
                Txn.on_completion() == OnComplete.UpdateApplication,
            ),
            Reject(),
        ],
    )

    return program


def clear_program():
    """clear program"""
    return Approve()


if __name__ == "__main__":
    with open("deposit_approval.teal", "w", encoding="utf-8") as f:
        COMPILED = compileTeal(
            approval_program(), mode=Mode.Application, version=6)
        f.write(COMPILED)

    with open("deposit_clear.teal", "w", encoding="utf-8") as f:
        COMPILED = compileTeal(
            clear_program(), mode=Mode.Application, version=6)
        f.write(COMPILED)



"""creates client and account"""
from algosdk import account, mnemonic


class Account:
    """user account"""

    def __init__(self, private_key=''):
        self.private_key = private_key
        self.public_key = account.address_from_private_key(private_key)

    def get_mnemonic(self):
        """returns mnemonic"""
        return mnemonic.from_private_key(self.private_key)

    @staticmethod
    def generate_account():
        """generates account"""
        return account.generate_account()



"""globals"""
from pyteal import Bytes, Int

CREATOR_KEY = Bytes("creator_key")

RESULT = Bytes("result")

TOKEN_FUNDING_KEY = Bytes("token_funding_key")
TOKEN_FUNDING_RESERVES = Bytes("token_funding_reserves")

POOL_FUNDING_RESERVES = Bytes("pool_funding_reserves")

POOL_TOKEN_KEY = Bytes("pool_token_key")
POOL_TOKENS_OUTSTANDING_KEY = Bytes("pool_tokens_outstanding_key")

YES_TOKEN_KEY = Bytes("yes_token_key")
YES_TOKENS_OUTSTANDING_KEY = Bytes("yes_tokens_outstanding_key")
YES_TOKENS_RESERVES = Bytes("yes_tokens_reserves")

NO_TOKEN_KEY = Bytes("no_token_key")
NO_TOKENS_OUTSTANDING_KEY = Bytes("no_tokens_outstanding_key")
NO_TOKENS_RESERVES = Bytes("no_tokens_reserves")

MIN_INCREMENT_KEY = Bytes("min_increment_key")

TOKEN_DEFAULT_AMOUNT = Int(10 ** 13)



"""pip package publishing"""
import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="algo_amm",
    version="0.0.2",
    author="pfedprog",
    author_email="pfedprog@gmail.com",
    description="Automated Prediction Market Maker on Algorand",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dspytdao/Algo_AMM",
    packages=setuptools.find_packages(),
    install_requires=['pyteal', 'py-algorand-sdk'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



"""algod client"""
from algosdk.v2client import algod
from algosdk.transaction import AssetConfigTxn


class AlgoClient:
    """algod client"""
    ALGOD_ADDRESS = "https://testnet-algorand.api.purestake.io/ps2"

    def __init__(self, algod_token):
        self.algod_token = algod_token
        self.headers = {
            "X-API-Key": algod_token,
        }
        self.client = algod.AlgodClient(
            self.algod_token, self.ALGOD_ADDRESS, self.headers)
        self.params = self.client.suggested_params()

    def wait_for_confirmation(self, tx_id):
        """
        Utility to monitor transaction confirmation.
        Args:
            tx_id: transaction id
        """
        last_round = self.client.status().get("last-round")
        tx_info = self.client.pending_transaction_info(tx_id)
        while not (tx_info.get("confirmed-round") and tx_info.get("confirmed-round") > 0):
            print("Waiting for confirmation...")
            last_round += 1
            self.client.status_after_block(last_round)
            tx_info = self.client.pending_transaction_info(tx_id)
        confirmed_round = tx_info.get("confirmed-round")
        print(f"Transaction {tx_id} confirmed in round {confirmed_round}.")
        return tx_info

    def create_asset(self, account):
        """
        Create asset.
        Args:
            account: account to sign and assign created asset
        """
        sender = account.public_key

        txn = AssetConfigTxn(
            sender=sender,
            sp=self.params,
            total=1_000_000_000,
            default_frozen=False,
            unit_name="AlgoAMM",
            asset_name="coin",
            manager=sender,
            reserve=sender,
            freeze=sender,
            clawback=sender,
            strict_empty_address_check=False,
            url=None,
            decimals=0)

        signed_txn = txn.sign(account.private_key)

        tx_id = self.client.send_transaction(signed_txn)
        print(f"Signed transaction with txID: {tx_id}")
        response = self.wait_for_confirmation(tx_id)
        print("TX ID: ", tx_id)
        confirmed_round = response['confirmed-round']
        print(f"Result confirmed in round: {confirmed_round}")
        return response['asset-index']
